extends Node2D
tool #allow rendering of selection box in the editor

onready var Grid = get_node("../Grid")

#currently selected item
var _currently_selecting = null
#current tile being hovered over
var _tile := Vector2.ZERO setget set_tile

var _height = 0


#this controls the cursors position
func set_tile(new_tile: Vector2) -> void:
	
	#this does all the moving checks, and returns null if invalid
	var target_position = Grid.move_cursor_request(new_tile, _height)
	
	#if null isn't returned then the cursor can be moved and tile set
	if target_position:
		position = target_position[0]
		_tile = new_tile
		_height = target_position[1]
		z_index = _height
		if (target_position[2] == false): #not behind
			z_index = z_index + 20
	
	#starts a timer so that the movement is "staggered" - limit on speed
	$Timer.start()


#this draws a 32x16 diamond in the center of the cursor
func _draw() -> void:
	var selection_color = Color.lightblue
	var selection_thickness = 2.0
	
	#I don't know how to draw use draw_polygon - hence lines
	draw_line(Vector2(-16,0), Vector2(0,8), selection_color, selection_thickness)
	draw_line(Vector2(16,0), Vector2(0,8), selection_color, selection_thickness)
	draw_line(Vector2(-16,0), Vector2(0,-8), selection_color, selection_thickness)
	draw_line(Vector2(16,0), Vector2(0,-8), selection_color, selection_thickness)


#this function I mainly took from a GDQuest tutorial - but with a few modifications
#it detects inputs for movement moves the cursor based off that
#it also handles clicks/enter logic for moving units
func _unhandled_input(event: InputEvent) -> void:
	#if the mouse is moved then change the selected tile (and therefore position) of the cursor
	if event is InputEventMouseMotion:
		self._tile = Grid.position_to_coordinates(event.position, 0)
	
	#if a left click/enter acton occurs then run the selection and movement logic
	#this bit of code resets the current selection after a move or if the selection was invalid
	# therefore requireing 2 successful selections to make a move (1st player, 2nd space)
	elif (event is InputEventMouseButton and event.pressed == false and
		event.button_index == BUTTON_LEFT or event.is_action_pressed("ui_accept")):
		
		#asks the selected node to to move the now selected tile if something has already been selected
		if _currently_selecting:
			var _currently_selecting_name = Grid.grid[_currently_selecting.x][_currently_selecting.y][0]
			get_node(("../" + _currently_selecting_name)).move(_tile - _currently_selecting)
			_currently_selecting = null
			Grid.remove_selection()
		
		#therefore, if nothing is currently selected, then we need to select what we have clicked on
		elif Grid.grid[_tile.x][_tile.y]:
			_currently_selecting = _tile
			var _currently_selecting_name = Grid.grid[_currently_selecting.x][_currently_selecting.y][0]
			if get_node("../" + _currently_selecting_name).movement_points != 0:
				Grid.display_selection(_tile, get_node("../" + _currently_selecting_name))
			else:
				_currently_selecting = null
		
		#if there was no unit under the cursor then do nothing (and reset to no selection)
		else:
			_currently_selecting = null
		
		get_tree().set_input_as_handled()
	
	#this adds the ability to undo the selection of a unit through right click or esc
	elif (event is InputEventMouseButton and event.pressed == false and
		event.button_index == BUTTON_RIGHT or event.is_action_pressed("ui_cancel")):
		
		_currently_selecting = null
		get_tree().set_input_as_handled()

	# The code below is for the cursor's movement.
	# The following lines make some preliminary checks to see whether the cursor should move or not
	# if the user presses an arrow key.
	var should_move := event.is_pressed()
	# If the player is pressing the key in this frame, we allow the cursor to move. If they keep the
	# keypress down, we only want to move after the cooldown timer stops.
	if event.is_echo():
		should_move = should_move and $Timer.is_stopped()

	# And if the cursor shouldn't move, we prevent it from doing so.
	if not should_move:
		return

	# Here, we update the cursor's current tile based on the input direction. See the set_tile()
	# function below to see what changes that triggers.
	if event.is_action("ui_right"):
		self._tile += Vector2.RIGHT
	elif event.is_action("ui_up"):
		self._tile += Vector2.UP
	elif event.is_action("ui_left"):
		self._tile += Vector2.LEFT
	elif event.is_action("ui_down"):
		self._tile += Vector2.DOWN

