extends Node

#maximum size needed for current resolution
const GRID_SIZE = Vector2(46 + 10,32 + 10)

#found these through trial and error
#these are actually the tile-size/2 (kinda)
const X_MATRIX = Vector2(16,8)
const Y_MATRIX = Vector2(-16,8)

#1/(Xx)(Yy)-(Yx)(Xy)
const INVERSE_MATRIX_MULTIPLIER = 1.0/((X_MATRIX.x*Y_MATRIX.y)-(Y_MATRIX.x*X_MATRIX.y))

#{ Xx Yx }		to		{ Yy -Yx }
#{ Xy Yy }		to		{ -Xy Xx }
const INVERSE_X_MATRIX = Vector2(Y_MATRIX.y,-X_MATRIX.y)
const INVERSE_Y_MATRIX = Vector2(-Y_MATRIX.x,X_MATRIX.x)

const EMPTY_TILE = [-1]
const GROUND_TILE = [0,1]
const WATER_TILE = [2]
const AIR_TILE = [5]
const LAVA_TILE = [4]
const OBSTACLE_TILE = []

#setget works here because the value is gotten after the object has been
# initalised (not before), so that the GRID_SIZE actually has values
onready var grid = _create_grid(GRID_SIZE.x, GRID_SIZE.y, true)
onready var TERRAIN_GRID = _create_grid(GRID_SIZE.x, GRID_SIZE.y)


func _create_grid (x, y, unpopulated : bool = false) -> Vector2:
	var out = []
	
	#creates an array with the range from 0 to x-1 (total being x)
	# with a second array inside each with the range 0 to y-1 (total being y)
	for _i in range(x):
		var col = []
		col.resize(y)
		out.append(col)
	
	if unpopulated:
		return out
	
	#populates the grid with the ids of the tilesets and the heights
	#the +1, +2 etc is so that the raised piece is set to the actual position is represents
	#this overrides only if the new tile wasn't empty - so the top tile and correct height are set 
	for i in range (GRID_SIZE.x):
		for j in range (GRID_SIZE.y):
			out[i][j] = [get_node("../TileMap0").get_cell(i,j), 0]
	for i in range (GRID_SIZE.x):
		for j in range (GRID_SIZE.y):
			if get_node("../TileMap1").get_cell(i,j) != EMPTY_TILE[0]:
				out[i+1][j+1] = [get_node("../TileMap1").get_cell(i,j), 1]
	for i in range (GRID_SIZE.x):
		for j in range (GRID_SIZE.y):
			if get_node("../TileMap2").get_cell(i,j) != EMPTY_TILE[0]:
				out[i+2][j+2] = [get_node("../TileMap2").get_cell(i,j), 2]
	for i in range (GRID_SIZE.x):
		for j in range (GRID_SIZE.y):
			if get_node("../TileMap3").get_cell(i,j) != EMPTY_TILE[0]:
				out[i+3][j+3] = [get_node("../TileMap3").get_cell(i,j), 3]
	
	return out


func register_new_unit (unit : Unit) -> Vector2:
	#turn the position into coordinates
	var current_coordinates = position_to_coordinates(unit.position, unit.height)
	
	#add the name (which must be unique due to godot) to the grid
	grid[current_coordinates.x][current_coordinates.y] = [unit.name, unit.height]
	
	#snaps the unit to the correct tile
	return coordinates_to_position(current_coordinates, unit.height)


func move_request (movement_coordinates : Vector2, unit : Unit): #TODO 3D
	# recieve a movement request,			DONE
	# check its valid,						MOSTLY DONE (HAPPY PATH!!!)
		# height difference check needed... somehow		TODO
		# change check existing tile check		TODO
		# check moving across water
	# change the 2d array,					DONE
	# send signals out to refresh position	CHANGED/DONE
		# now just returns the value for the unit to move 

	var current_coordinates = _find_grid_coordinates(unit.name)
	var target_coordinates = current_coordinates + movement_coordinates
	var target_co_height = TERRAIN_GRID[target_coordinates.x][target_coordinates.y][1]
	
	var area_around = get_around_coordinate(current_coordinates, unit.movement_points)
	var area_moveable = _filter_moveable_spaces(area_around, unit.water_walk, unit.air_walk, unit.lava_walk)
	
	if not area_moveable.has(target_coordinates):
		return null
	
	#check if an another unit is in that space
	if grid[target_coordinates.x][target_coordinates.y] != null:
		return null
	
	#check if its totally behind(invalid) or partially behind(changes z_index)
	#this slightly overlaps the filtering code, but not by much
	var behind = false
	if (get_tile_behind_state(target_coordinates) == "d"):
		return null
	elif(get_tile_behind_state(target_coordinates) == "p"):
		 behind = true

	#update where that unit is on the grid
	_move_unit_in_grid(current_coordinates, target_coordinates)

	#return a value which is moved too
	return [coordinates_to_position(target_coordinates, target_co_height), target_co_height, behind]


func move_cursor_request (target_coordinates : Vector2, height : int): #TODO 3D - no clue what to do with the cursor
	#Check if its moving outside the grid
	if (not _is_in_grid(target_coordinates)):
		return null
	
	var behind = false
	if (get_tile_behind_state(target_coordinates) == "d"):
		return null
	elif(get_tile_behind_state(target_coordinates) == "p"):
		 behind = true
	
	height = TERRAIN_GRID[target_coordinates.x][target_coordinates.y][1]
	
	#Return a value which is moved too
	#The tile_size/2 is because nodes are in the center of a tile
	return [coordinates_to_position(target_coordinates, height), height, behind]


func _find_grid_coordinates (name : String):
	for x in GRID_SIZE.x:
		for y in GRID_SIZE.y:
			if grid[x][y]:
				if grid[x][y][0] == name:
					return Vector2(x,y)
	return null


func _move_unit_in_grid (from : Vector2, to : Vector2): #TODO 3D
	#this is a simple method that moves one value to another place on the grid
	var value = grid[from.x][from.y]
	grid[from.x][from.y] = null
	grid[to.x][to.y] = value


func coordinates_to_position (coordinates : Vector2, height: int) -> Vector2:
	var out = Vector2(0,-240) #the position of the ../TileMap 0,0
	out += (coordinates.x) * X_MATRIX #add the x
	out += (coordinates.y) * Y_MATRIX #add the y
	
	out.y = out.y - height * 16
	return out


func position_to_coordinates (node_position : Vector2, height : int) -> Vector2:
	#reverse matrix.... fun
	var out = node_position - Vector2(0,-240)
	var tmp_x = out.x
	out.x = INVERSE_X_MATRIX.x*INVERSE_MATRIX_MULTIPLIER*out.x + INVERSE_Y_MATRIX.x*INVERSE_MATRIX_MULTIPLIER*out.y
	out.y = INVERSE_X_MATRIX.y*INVERSE_MATRIX_MULTIPLIER*tmp_x + INVERSE_Y_MATRIX.y*INVERSE_MATRIX_MULTIPLIER*out.y
	
	#round the decimal places. Not sure why the minus one - but its needed
	#each addition in height reduces each coordinate by one, so that needs to be
	# added back
	out.x = round(out.x) + height
	out.y = round(out.y) + height
	
	return out


func _is_in_grid (coordinates : Vector2) -> bool:
	if (coordinates.x >= GRID_SIZE.x or
		coordinates.y >= GRID_SIZE.y or
		coordinates.x < 0 or
		coordinates.y < 0):
		return false
	return true


func get_tile_behind_state (coordinates : Vector2) -> String:
	#checking if tiles are directly behind
	#the plus 1 and plus 2 mean that its a height difference of 2 or 3 required respectively 
	if (TERRAIN_GRID[coordinates.x + 1][coordinates.y + 1][1] > TERRAIN_GRID[coordinates.x][coordinates.y][1] or
		TERRAIN_GRID[coordinates.x + 2][coordinates.y + 2][1] > TERRAIN_GRID[coordinates.x][coordinates.y][1] + 1 or
		TERRAIN_GRID[coordinates.x + 3][coordinates.y + 3][1] > TERRAIN_GRID[coordinates.x][coordinates.y][1] + 2):
		return "d" #directly
	
	#checking if tiles are partially covered
	elif(TERRAIN_GRID[coordinates.x][coordinates.y + 1][1] > TERRAIN_GRID[coordinates.x][coordinates.y][1] or
		TERRAIN_GRID[coordinates.x + 1][coordinates.y][1] > TERRAIN_GRID[coordinates.x][coordinates.y][1] or
		TERRAIN_GRID[coordinates.x + 2][coordinates.y + 1][1] > TERRAIN_GRID[coordinates.x][coordinates.y][1] + 1 or
		TERRAIN_GRID[coordinates.x + 1][coordinates.y + 2][1] > TERRAIN_GRID[coordinates.x][coordinates.y][1] + 1 or
		TERRAIN_GRID[coordinates.x + 3][coordinates.y + 2][1] > TERRAIN_GRID[coordinates.x][coordinates.y][1] + 2 or
		TERRAIN_GRID[coordinates.x + 2][coordinates.y + 3][1] > TERRAIN_GRID[coordinates.x][coordinates.y][1] + 2):
		return "p" #partially
	
	else:
		return "n" #no


func get_around_coordinate (coordinate : Vector2, movement_distance : int):
	#found https://godotengine.org/qa/64496/how-to-get-a-random-tile-in-a-radious
	
	#disclude the tile being stood on
	movement_distance = movement_distance + 1
	
	#gets the rectangle bounding the circle
	var radius_vec = Vector2(movement_distance, movement_distance)
	var min_pos = (coordinate - radius_vec).floor()
	var max_pos = (coordinate + radius_vec).ceil()

	#convert the mins/maxs to integers so we can use range for loop
	var min_x = int(min_pos.x)
	var max_x = int(max_pos.x)
	var min_y = int(min_pos.y)
	var max_y = int(max_pos.y)

	var positions = []

	#gathers all the points that are within the radius
	for y in range(min_y, max_y):
		for x in range(min_x, max_x):
			var tile_pos = Vector2(x, y)
			if tile_pos.distance_to(coordinate) < movement_distance:
				 positions.append(tile_pos)

	return positions


func _filter_moveable_spaces (positions_array, water_walk : bool = false, air_walk : bool = false, lava_walk : bool = false):
	#this uses happy path to filter out the invalid move spaces
	var new_pos_array = []
	for i in positions_array.size():
		var tile_type = TERRAIN_GRID[positions_array[i].x][positions_array[i].y][0]
		
		if (get_tile_behind_state(positions_array[i]) == "d"):
			continue
		
		if (EMPTY_TILE.has(tile_type) or OBSTACLE_TILE.has(tile_type)):
			continue
		
		if (WATER_TILE.has(tile_type) and not water_walk):
			continue
		
		if (AIR_TILE.has(tile_type) and not air_walk):
			continue
		
		if (LAVA_TILE.has(tile_type) and not lava_walk):
			continue
		
		new_pos_array.append(positions_array[i])
	return new_pos_array


func display_selection (coordinate : Vector2, unit : Unit):
	var movement_amount = unit.movement_points
	print("hi")
	for i in range (movement_amount, -1, -1):
		var area_around = get_around_coordinate(coordinate, i)
		var area_around_smaller = get_around_coordinate(coordinate, i -1)
		if i == 1:
			area_around_smaller = []
		
		for j in area_around_smaller.size():
				if area_around.has(area_around_smaller[j]):
					area_around.remove(area_around.find(area_around_smaller[j]))
		
		var area_moveable = _filter_moveable_spaces(area_around, unit.water_walk, unit.air_walk, unit.lava_walk)
		$SelectionAreaController.create_selection_from_array(area_moveable, i)


func remove_selection ():
	$SelectionAreaController.emit_signal("remove_selection")








	#func get_around_coordinate(coordinate : Vector2, movement_distance : int):
#	#this generates the list of coordinates that a unit can move too
#	#the radius + radius (either side) + 1 (the center) 
#	var positions_array = []
#	for i in movement_distance + movement_distance + 1:
#		for j in movement_distance + movement_distance + 1:
#			positions_array.append(coordinate - Vector2(movement_distance, movement_distance) + Vector2(i,j))
#
#	return positions_array
